/** (c) Brendan Macmillan 2002, licensed under GNU's GPL 2 
		(If your project is not GPL, then a fee is due - see license on website)
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@objectlight.com */


package JSX.magic;
import sun.misc.Unsafe;
import java.io.*;
import java.lang.reflect.*;

import sun.reflect.ReflectionFactory;
import java.security.AccessController;
import java.security.PrivilegedAction;

class MagicClass14 extends MagicClassI {
	static ReflectionFactory reflFactory = (ReflectionFactory)
		AccessController.doPrivileged(
			new ReflectionFactory.GetReflectionFactoryAction());
	static Constructor objectConstructor;
	static Unsafe unsafe;
	public MagicClass14() {
		// allocate new instance: init Serialization Constructor
		try {
			objectConstructor = Object.class.getDeclaredConstructor(new Class[0]);
		} catch (NoSuchMethodException noMethod) {
			throw new InternalError("Object constructor not found");
		}

		// set final fields: init Unsafe
		try {
			Class osc$frClazz = Class.forName("java.io.ObjectStreamClass$FieldReflector");
			Field unsafeF = osc$frClazz.getDeclaredField("unsafe");
			unsafeF.setAccessible(true);
			unsafe = (Unsafe) unsafeF.get(null);	//static field
		} catch (ClassNotFoundException noClass) {
			throw new InternalError("java.io.ObjectStreamClass$FieldReflector not found");
		} catch (NoSuchFieldException noField) {
			throw new InternalError("java.io.ObjectStreamClass$FieldReflector has no field 'unsafe'");
		} catch (IllegalAccessException noAccess) {
			throw new InternalError("not able to access java.io.ObjectStreamClass$FieldReflector.unsafe");
		}

	// set osc (note: different Exceptions from other magics)
		try {
			osc = (ObjectStreamClass) newInstance(ObjectStreamClass.class);
		} catch (IllegalAccessException noAccess) {
			throw new InternalError(noAccess.toString());
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		} catch (NotSerializableException e) {	// Serializable check in newInstance
			e.printStackTrace();
		}
	}



//Experiment with demanding classes implement "Serializable" - see similar
//code in MagicClass13.newInstance()
	public Object newInstance(Class currentClass)
			throws InvocationTargetException,
						 NotSerializableException,
			       IllegalAccessException {
/*
		if (!Serializable.class.isAssignableFrom(currentClass))
			throw new NotSerializableException("'"+currentClass+"' is not Serializable - JSX *can* serialize such classes, despite them not implementing the 'Serializable' interface; but this feature is currently switched off (since version JSX1.0.4.2), as an experiment to see if it makes any difference.  If this is an annoyance to you, please post to the mailing list immediately:\n" +
			"  http://groups.yahoo.com/group/JSX-ideas/messages");
*/

		
		Constructor clCons = null;
		// Serializable -> JOS way (inherited non-Serializable different)
		if (Serializable.class.isAssignableFrom(currentClass)) {
			clCons = getSerializableConstructor(currentClass);
			if (clCons==null)
				throw new InternalError("Unable to find no-arg constructor of first non-serializable subclass of "+currentClass);
		} else {
		// !Serializable -> JSX way (just does it)
			clCons = reflFactory.newConstructorForSerialization(currentClass, objectConstructor);
			clCons.setAccessible(true);
		}

		Object obj = null;
		try {
			obj = clCons.newInstance(new Object[0]);
		} catch (InstantiationException cantMake) {
			throw new InternalError("serialization constructor for "+currentClass+" not not able to instantiate it");
		}
		return obj;
	}

    /**
     * Returns subclass-accessible no-arg constructor of first non-serializable
     * superclass, or null if none found.  Access checks are disabled on the
     * returned constructor (if any).
     */
	private static Constructor getSerializableConstructor(Class cl) {
		Class initCl = cl;
		while (Serializable.class.isAssignableFrom(initCl)) {
			if ((initCl = initCl.getSuperclass()) == null) {
				return null;
			}
		}
		try {
			Constructor cons = initCl.getDeclaredConstructor(new Class[0]);
			int mods = cons.getModifiers();
			if ((mods & Modifier.PRIVATE) != 0 ||
				((mods & (Modifier.PUBLIC | Modifier.PROTECTED)) == 0 &&
				!packageEquals(cl, initCl))) {
				return null;
			}
			cons = reflFactory.newConstructorForSerialization(cl, cons);
			cons.setAccessible(true);
			return cons;
		} catch (NoSuchMethodException ex) {
			return null;
		}
	}

    /**
     * Returns true if classes are defined in the same package, false
     * otherwise.
     */
	private static boolean packageEquals(Class cl1, Class cl2) {
		Package pkg1 = cl1.getPackage(), pkg2 = cl2.getPackage();
		return ((pkg1 == pkg2) || ((pkg1 != null) && (pkg1.equals(pkg2))));
	}



//Exception back-compatible - good to rewrite JSX, so caller can handle nicer
	public void setObjectFieldValue(Object parent, Field f, Class type, Object value) {
			//if (parent==null) throw new NullPointerException();
		if (value!=null && !f.getType().isInstance(value))//if non-null & incompat
			throw new ClassCastException("Expected "+f.getType()+" or subclass, got "+value.getClass());
		unsafe.putObject(parent, unsafe.fieldOffset(f), value);
	}

	public void setPrimitiveFieldValues(Object parent, Field f, Object value) {
			//if (parent==null) throw new NullPointerException();
		int key = unsafe.fieldOffset(f);
		Class type = f.getType();	//value.getClass() gives Integer.class etc
		if (type==int.class)	//ordered by expected frequency of occurance
			unsafe.putInt(parent, key, ((Integer)value).intValue());
		else if (type==double.class)
			unsafe.putDouble(parent, key, ((Double)value).doubleValue());
		else if (type==boolean.class)
			unsafe.putBoolean(parent, key, ((Boolean)value).booleanValue());
		else if (type==byte.class)
			unsafe.putByte(parent, key, ((Byte)value).byteValue());
		else if (type==char.class)
			unsafe.putChar(parent, key, ((Character)value).charValue());
		else if (type==short.class)
			unsafe.putShort(parent, key, ((Short)value).shortValue());
		else if (type==float.class)
			unsafe.putFloat(parent, key, ((Float)value).floatValue());
		else if (type==long.class)
			unsafe.putLong(parent, key, ((Long)value).longValue());
		else
			throw new InternalError(type+" is unknown type, expected primitive");
	}
}
